/**
\file neighborhood.hpp
\date 17/04/2021
\author Eugene Pin
\version 1
\brief Neighborhood

Cette classe représente le voisinage d'une cellule.

**/

#ifndef NEIGHBOURHOOD_HPP
#define NEIGHBOURHOOD_HPP

#include <map>
#include <utility>
#include <string>
#include "coord.hpp"
#include "state.hpp"

using namespace std;

//! \brief Exception lancée lors d'une erreur de manipulation d'un voisinage.
class NeighborhoodException : public std::exception
{
    std::string _msg;
public:
    NeighborhoodException(const std::string& msg) : _msg(msg){}

    virtual const char* what() const noexcept override
    {
        return _msg.c_str();
    }
};

//! \brief Classe représentant un voisinage, c'est à dire une liste de cellules dont leur état et leur position.
class Neighborhood {
    //! \brief Map de voisins contenant la position relative et l'état du voisin
    map<Coord, unsigned int> neighborPositions;


public:
    //! \brief Fonction vérifiant que la coordonnée d'un voisin que l'on souhaite ajouter dans le vecteur neighborPositions est unique
    //! \param Coord : Coordonnée à comparer
    //! \return retourne vrai si la coordonnée est unique sinon faux s'il existe un doublon
    bool isUnique(Coord c) const;

	//! \brief Retourne le nombre de voisin ayant l'état définit en paramètre
	//! \param State : Etat des voisins à rechercher
	//! \return Le nombre de voisin
    unsigned int getNb(unsigned int s) const;

	//! \brief Retourne l'état de la cellule voisine située à la coordonnée relative entrée en paramètre
	//! \param Coordonnée du voisin à trouver
	//! \return L'état de la cellule
    unsigned int getAt(Coord c) const;

	//! \brief Ajoute un voisin au vecteur des voisins. Prend une coordonnée et un état en paramètre
	//! \param Coord, State : coordonée relative et état du voisin
    void addNeighbor(Coord c, unsigned int s) {
		if(isUnique(c))
            neighborPositions[c] = s;
        else
            throw NeighborhoodException("La coordonnée existe déjà. \n");
	}

    //! \brief Retourne une copie du voisinage, miroir selon l'axe vertical.
    Neighborhood flip_vertically() const;
    //! \brief Retourne une copie du voisinage, miroir selon l'axe horizontal.
    Neighborhood flip_horizontally() const;
    //! \brief Retourne une copie du voisinage, tournée de 90° dans le sens des aiguilles d'une montre.
    Neighborhood rotate90() const;

    //! \brief Retourne le nombre de voisins dans le Neighborhood.
    unsigned size() const
    { return neighborPositions.size(); }

    //! \brief Retourne la cellule numéro i (index allant de 0 à size()-1, permettant d'indexer les cellules indépendamment de leur position
    pair<Coord, unsigned int> neighbor_at_index(unsigned i) const
    {
        if (i >= size())
            throw NeighborhoodException("Index invalide\n");
        auto it = neighborPositions.begin();
        std::advance(it, i);
        return *it;
    }
};

#endif // NEIGHBOURHOOD_HPP
